;nyquist plug-in
;version 3
;type process
;categories "http://lv2plug.in/ns/lv2core#FunctionPlugin"
;name "Punch Copy/Paste..."
;action "Working..."
;info "by Steve Daulton. (www.easyspacepro.com).\nReleased under GPL v2.\n"

;; PunchCopy.ny by Steve Daulton March 2013.
;; Released under terms of the GNU General Public License version 2:
;; http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
;;
;; For information about writing and modifying Nyquist plug-ins:
;; http://wiki.audacityteam.org/wiki/Nyquist_Plug-ins_Reference


;control mode "Select action" choice "Copy (safe),Copy (replace),Paste,Clear Clipboard" 0
;control loop "Paste: Loop mode" choice "Enable looping,Paste once only" 0
;control curve "Crossfade type" choice "Linear,S-Curve,Eq Power" 0
;control fade-seconds "Max crossfade time (s)" real "" 0 0 5
;control mix "Paste mix (0 - 100 %)" int "" 100 0 100

(setq id "PunchCopyPaste_19-03-2013_01")
(setq mix (max 0 (min 1 (/ mix 100.0))))

;;; Generate property symbol names
(defun propgen (prop id)
  (eval (read (make-string-input-stream 
    (format nil "'~a_~a" prop id)))))

;; clipboard data
(setq left (propgen "left" id))             ; left/mono channel
(setq right (propgen "right" id))           ; right channel
(setq clipdur (propgen "clipdur" id))       ; clipboard duration (seconds)
(setq cliplen (propgen "cliplen" id))       ; clipboard length (samples)
(setq clipsrate (propgen "clipsrate" id))   ; clipboard sample rate

;; paste parameters
(setq looptype (propgen "looptype" id))     ; loop or one-shot
(setq fadetype (propgen "fadetype" id))     ; linear, s-curve or cosine
(setq fadedur (propgen "fadedur" id))       ; crossfade duration (seconds)
(setq pastemix (propgen "pastemix" id))     ; mix ratio

;;; Save settings for PunchPaste.
(defun save-settings ()
  (putprop '*scratch* loop looptype)          ; loop or one-shot
  (putprop '*scratch* curve fadetype)         ; linear, s-curve or cosine
  (putprop '*scratch* fade-seconds fadedur)   ; crossfade duration (seconds)
  (putprop '*scratch* mix pastemix))          ; mix ratio

;;; Clear properties from *scratch*.
(defun clear ()
  (clear-scratch (list
    left      right       clipdur
    cliplen   clipsrate   looptype
    fadetype  fadedur     pastemix))
  "Clipboard cleared.")

(defun clear-scratch (proplist)
  (dotimes (i (length proplist))
    (remprop '*scratch* (nth i proplist))))

;;; Copy to clipboard
(defun copy (sig)
  (let ((s (multichan-expand #'extract-abs 0 (/ 100000000 *sound-srate*) s)))
      (if (arrayp sig)
          (progn  ; stereo track
            (putprop '*scratch* (aref s 0) left)
            (putprop '*scratch* (aref s 1) right))
          ; else left/mono
          (progn
            (putprop '*scratch* s left)
            (remprop '*scratch* right)))
    (putprop '*scratch* (get-dur s) clipdur)
    (putprop '*scratch* (min 100000000 len) cliplen)
    (putprop '*scratch* *sound-srate* clipsrate)
    ; must reference sound in output so that sound in *scratch* is not cleared
    (format nil "~a copied.~a" 
            (ss-to-mmss
              (if (arrayp s)
                  (aref (multichan-expand #'get-dur s) 1)
                  (get-dur s))
              1 3)
            (if (= mix 0)
                (format nil "~%~%Warning: 'Paste Mix' is set to 0%.~%~
                  Set 'Paste Mix' to 100% to replace the selected audio~%~
                  with the clipboard contents, or between 0 and 100%~%~
                  for a mix of both.")
                ""))))

;;; Crossfade clipboard and seletion
(defun mixpaste (sig)
;; for single shot we want the fade out at the end of the 
;; clipboard  (clipdur-rel) unless the selection is shorter,
;; in which case we want the fade-out at the end of the selection (1).
;; For loop paste we want the selection length (1).
  (let* ((cdur (if (= loop 1) (min (clipdur-rel) 1) 1))
         ; fade duration local time (min 50%)
         (fdur (min (* 0.5 cdur) (/ (get '*scratch* fadedur)(get-duration 1))))
         ; get clip to paste
         (audioclip (paste cdur))
         ; Fade in/out for audioclip
         (env (make-env cdur fdur curve mix))
         ;; Fade out/in for selection
         (ienv 
           (if (= curve 2)
               (if (< (get '*scratch* fadedur)(/ 4 *sound-srate*))
                   (diff 1 env)
                   (make-env cdur fdur 3 mix))
               (diff 1 env))))
    (sum
      (mult audioclip env)
      (mult sig ienv))))


(defun make-env (cdur fdur curve gain)
  (control-srate-abs *sound-srate*
    (if (< (get '*scratch* fadedur)(/ 4 *sound-srate*))
        (pwl fdur gain (- cdur fdur) gain cdur 0 1)
        (case curve
          ; linear fade
          (0 (pwl fdur gain (- cdur fdur) gain cdur 0 1))
          ;; Raised cosine (equal amplitude)
          (1 (mult gain
                (sim (s-rest 1)
                  (rcos (get-duration fdur) 0)
                  (if (> cdur (* 2 fdur))
                      (pwl fdur 0 fdur 1 (- cdur fdur) 1 (- cdur fdur) 0 cdur 0 1)
                      (s-rest 0))
                  (at (- cdur  fdur) (cue (rcos (get-duration fdur) 1))))))
          ;; 'Equal power' cosine curve.
          (2 (mult gain
                (sim (cosine (get-duration fdur) 0)
                  (if (> cdur (* 2 fdur))
                      (pwl fdur 0 fdur 1 (- cdur fdur) 1 (- cdur fdur) 0 cdur 0 1)
                      (s-rest 1))
                  (at (- cdur  fdur) (cue (cosine (get-duration fdur) 1))))))
          ;;  Inverse 'Equal power' cosine curve.
          (T (sum (- 1 gain)
              (mult gain
                (sim (cosine (get-duration fdur) 1)
                  (pwl cdur 0 cdur 1 1 1)
                  (at (- cdur  fdur)
                      (cue (cosine (get-duration fdur) 0)))))))))))

;;; 'Equal power' cosine curve.
(defun cosine (dur phase)
; phase=0 for fade-in. phase=1 for fade out
  (let ((freq (hz-to-step (/ 0.25 dur))))
    (abs-env
      (osc freq dur *sine-table* (* phase 90)))))

;;; Raised cosine 'S-curve'
(defun rcos (dur phase)
; phase=0 for fade-in. phase=1 for fade out
  (let ((phase (if (= phase 0) -90 90))
        (freq (hz-to-step (/ 0.5 dur))))
    (abs-env
      (mult 0.5
        (sum 1 
          (osc freq dur *sine-table* phase))))))

;;; Prepare audio for pasting
(defun paste (duration)
  (if (arrayp s)  ; stereo output
      (if (get '*scratch* right) ; stereo on clipboard
          (PasteStereo duration)
          (let ((channel (PasteMono duration)))
            (vector channel channel)))
      ;; mono output
      (if (get '*scratch* right) ; stereo on clipboard
          (let ((output (PasteStereo duration)))
            (scale 0.5 
              (sum (aref output 0)(aref output 1))))
          (PasteMono duration))))

(defun PasteMono (dur)
  (let ((repeats (ceil (/ (get-duration dur)(get '*scratch* clipdur)))))
    (setf clip (seqrep (val repeats)(get '*scratch* left)))
    (when (/= *sound-srate* (get '*scratch* clipsrate))
      (setf clip (resample clip *sound-srate*)))
    (extract 0 1 clip)))

(defun PasteStereo (dur)
  (let ((repeats (ceil (/ (get-duration dur)(get '*scratch* clipdur)))))
    (setf leftclip (seqrep (val repeats)(get '*scratch* left)))
    (setf rightclip (seqrep (val repeats)(get '*scratch* right)))
    (when (/= *sound-srate* (get '*scratch* clipsrate))
      (setf leftclip (resample leftclip *sound-srate*))
      (setf rightclip (resample rightclip *sound-srate*)))
    (vector
      (extract 0 1 leftclip)
      (extract 0 1 rightclip))))


;;; Duration of clipboard as proportion of selection
(defun clipdur-rel ()
  (let ((sr-ratio (/ *sound-srate* (get '*scratch* clipsrate)))
        (clipdur (get '*scratch* clipdur)))
    (/ (* clipdur sr-ratio)(get-duration 1))))

;;; Absolute duration of clipboard (max 100 million samples)
(defun get-dur (sig)
  (if (arrayp sig)
    (/ (snd-length (aref sig 0) 100000000) *sound-srate*)
    (/ (snd-length sig 100000000) *sound-srate*)))

;;; ceiling function
(defun ceil (x)
  (if (> x (truncate x))
      (1+ (truncate x))
      (truncate x)))

;;; seconds to mm:ss.x
(defun ss-to-mmss (ss round places)
; round 0 = truncate, 1 = round, 2 = round up.
; places = number of decimal places.
  (setq ss (float ss))
  (let* ((mm (truncate (/ ss 60.0)))
         (dec (- ss (truncate ss)))
         (ss (rem (truncate ss) 60))
         (dplaces (power 10 places)))
    (setq dec
      (case round
        (0 (/ (truncate (* dec dplaces)) dplaces))
        (1 (/ (round (* dec dplaces)) dplaces))
        (T (/ (ceil (* dec dplaces)) dplaces))))
    (setq ss (+ ss dec))
    (setf *float-format* (format nil "%.~af" places))
    (format nil "~a~a~a s"
           (if (> mm 0) mm "")
           (if (> mm 0) " m " "")
           ss)))


;; Error checking and process
(cond
  ; len is negative if sample count overflows
  ((and (or (= mode 0)(= mode 1))
        (or (< len 1)(> len 100000000)))
    (format nil
      "Error.~%~
      Maximum length that can be copied is 100 million~%~
      samples (~a seconds at ~a Hz sample rate)."
      (ss-to-mmss (/ 100000000 *sound-srate*) 0 1)
      (truncate *sound-srate*)))
  ((and (get '*scratch* left)(= mode 0))
    (format nil
      "Error.~%~
      'Copy (safe)' protects the clipboard from being ~
      overwritten. Select action 'Clear Clipboard' ~
      to clear the clipboard, or use 'Copy (replace)' ~
      to overwrite the clipboard with the current ~
      selection."))
  ((and (= mode 2)
    (or (not (get '*scratch* left))
        (not (get '*scratch* clipdur))))
          "Error.\nClipboard empty.\nUse 'Copy' action first.")
  ((and (= mode 2)(= mix 0))
    (save-settings)
    (format nil "Settings saved.~%~
      Mix set to 0% so nothing to paste.~%~%~
      Set 'Paste Mix' to 100% to replace the selected audio~%~
      with the clipboard contents, or between 0 and 100%~%~
      for a mix of both."))
  (T  ; no errors detected so process
    (case mode
      ((0 1) (save-settings)
             (copy s))
      (2 (save-settings)
         (mixpaste s))
      (T (clear)))))